"""Function(s) for getting a format string from a date string."""

import re
from copy import copy
from itertools import product
from collections import defaultdict, Counter
from typing import Dict, Iterator, List
from dateutil.parser import parse, ParserError  # type: ignore

FORMAT_DIRECTIVES = [
    '%a', '%A', '%w', '%d', '%b', '%B', '%m', '%y', '%Y',
    '%H', '%I', '%p', '%M', '%S', '%f', '%z', '%Z', '%j',
    '%U', '%W', '%c', '%x', '%X', '%%',
    '%-d', '%-m', '%-H', '%-I', '%-M', '%-S', '%-j']

DATE_TOKEN_REGEX = re.compile(r'([\w]+)|([\W]+)')


# pylint: disable=too-many-locals,too-many-branches
def _get_all_format_strings(date_string, allow_reuse=False,
                            include_day_of_year=False,
                            include_week_number=False,
                            no_zero=False) -> Iterator[str]:
    """Yield date format strings which match date_string."""
    try:
        parsed_datetime = parse(date_string)
    except ParserError:
        yield ''
        return
    tokens = DATE_TOKEN_REGEX.findall(date_string)
    possible_matches = defaultdict(list)  # type: Dict[str, List[str]]
    format_directives = copy(FORMAT_DIRECTIVES)
    if not include_day_of_year:
        format_directives.remove('%-j')
        format_directives.remove('%j')
    if not include_week_number:
        for directive in ('%W', '%w', '%U'):
            format_directives.remove(directive)
    for test_format in format_directives:
        test_string = parsed_datetime.strftime(test_format)
        possible_matches[test_string.lower()].append(test_format)
        possible_parts = []
        for (token, delimiter) in tokens:
            if token:
                if token.lower() in possible_matches:
                    possible_parts.append(
                        possible_matches[token.lower()])
            else:
                possible_parts.append([delimiter])

        possible_combinations = list(product(*possible_parts))
        if not no_zero:
            possible_combinations = _remove_extra_nlz(  # type: ignore
                possible_combinations)
        for sequence in possible_combinations:
            if not allow_reuse:
                counts = Counter(sequence)
                max_count = max(
                    counts[diretive] for diretive in format_directives)
                if max_count > 1:
                    continue
            format_string = "".join(sequence)
            if parsed_datetime.strftime(
                    format_string).lower() == date_string.lower():
                yield format_string


def _remove_extra_nlz(combinations) -> Iterator[List[str]]:
    """Remove unnecessary non-leading-zero variants.

    Remove non-leading-zero directives (e.g. "%-I") if the corresponding
    leading-zero directive (e.g. "%I") is also a match.
    """
    for variant in combinations:
        is_duplicate = False
        for i, token in enumerate(variant):
            if '-' in token:
                with_zero = (variant[:i] +
                             (token.replace('-', ''),) +
                             variant[(i + 1):])
                if with_zero in combinations:
                    is_duplicate = True
                    break
        if not is_duplicate:
            yield variant


def get_unique_format_strings(date_string, allow_reuse=False) -> List[str]:
    """Return date format strings matching date_string, without duplicates."""
    return list(set(_get_all_format_strings(date_string, allow_reuse)))
